/*
 *  PicoWeb NE2000 Ethernet Controller Chip Test
 *
 *  To compile for PicoWeb:
 *
 *  avr-gcc -I %PWDEV%/include -O4 -c -Wa,-alms=isa.lst -o isa.o isa.c
 *
 */

#include "test.h"
#include "isa.h"
#include "ne2000cr.h"

void led_on(void);
void led_off(void);

#define NE2000_TYPE "NE1000"      /* 8-bit data bus */

static uchar err = 0;

int putchar(char ch)
{
    /*
     * Note: The PicoWeb AVR assembly language routine "do_putchar" prints
     * out the character passed to it in register r25.  It preserves all of
     * the AVR registers that it uses and therefore call be called directly
     * from C-code.  Because the machine code from the gcc compiler passes
     * int parameters in registers, starting with the register pair r24/r25,
     * "do_putchar" can be called directly from C-code.
     */
    if (ch == '\n') do_putchar('\r'<<8);
    do_putchar(ch<<8);
    return ch;
}

#define PROGMEM __attribute__ ((progmem))
#define PSTR(s) ({static char c[] PROGMEM = s;c;})
#define __lpm_macro(addr) ({                    \
        char t;                                 \
        asm volatile (                          \
                "lpm" "\n\t"                    \
                "mov    %0,r0"                  \
                : "=r" (t)                      \
                : "z" (addr)                    \
        );                                      \
        t;                                      \
})

#define PMPUTSTR(s) pmputstr(PSTR(s))

void pmputstr(unsigned char PROGMEM *s)
{
    unsigned char c ;

    while ((c = __lpm_macro(s++)) != 0) {
        putchar(c);
    }
}

void wait(void)
{
    PMPUTSTR("Press key to repeat...");
}

int getchar(void)
{
    while (com_in == com_out) ;
    return cser_getc();
}

void putstr(char *s)
{
    while (*s) {
        putchar(*s++);
    }
}

static void _putint(int i)
{
    if (i != 0) {
        _putint(i/10);
        putchar((i%10) + '0');
    }
}

void putint(int i)
{
    if (i == 0) {
        putchar('0');
    } else if (i < 0) {
        putchar('-');
        _putint(-i);
    } else {
        _putint(i);
    }
}

void cr(void)
{
    putchar('\r');
}

void crlf(void)
{
    putchar('\n');
}

void space(void)
{
    putchar(' ');
}

static void _puthexn(uchar i)
{
    if (i < 10) {
        putchar(i + '0');
    } else {
        putchar(i - 10 + 'a');
    }
}

void puthexb(uchar i)
{
    _puthexn(i>>4);
    _puthexn(i&15);
}

void puthex(ushort i)
{
    puthexb(i>>8);
    puthexb(i&0xff);
}

uchar DMAin(ushort addr)
{
    ISAout(neCRwr, neCR_STA | neCR_RD2 | neCR_STP);
    ISAout(neRBCR1wr, 0x00);
    ISAout(neRBCR0wr, 0x01);
    ISAout(neRSAR1wr, addr>>8);
    ISAout(neRSAR0wr, addr);
    ISAout(neCRwr, neCR_STA | neCR_RD0 | neCR_STP);
    return ISAin(neDMArw);
}

ushort DMAinw(ushort addr)
{
    ISAout(neCRwr, neCR_STA | neCR_RD2 | neCR_STP);
    ISAout(neRBCR1wr, 0x00);
    ISAout(neRBCR0wr, 0x01);
    ISAout(neRSAR1wr, addr>>8);
    ISAout(neRSAR0wr, addr);
    ISAout(neCRwr, neCR_STA | neCR_RD0 | neCR_STP);
    return ISAinw(neDMArw);
}

void DMAout(ushort addr, uchar val)
{
    ISAout(neCRwr, neCR_STA | neCR_RD2 | neCR_STP);
    ISAout(neRBCR1wr, 0x00);
    ISAout(neRBCR0wr, 0x01);
    ISAout(neRSAR1wr, addr>>8);
    ISAout(neRSAR0wr, addr);
    ISAout(neCRwr, neCR_STA | neCR_RD1 | neCR_STP);
    ISAout(neDMArw, val);
}

void Error(void)
{
    PMPUTSTR("Error");
}

void NE2000(void)
{
    PMPUTSTR(NE2000_TYPE);
}

void bang(void)
{
    putchar('!');
}

void oops(ushort addr, uchar val, uchar expect)
{
    cr();
    space();
    Error();
    bang();
    space();
    puthex(addr);
    putchar(':');
    puthexb(val);
    bang();
    putchar('=');
    puthexb(expect);
    crlf();
    err = 1;
}

void blank(void)
{
    uchar i = 72;

    cr();
    while (i--) space();
    cr();
}

void progress(ushort ram, uchar val, uchar mask)
{
    if (mask == 0 || (ram & mask) == 0) {
        cr();
        puthex(ram);
        putchar(':');
        puthexb(val);
    }
}

void show_regs(void) {
    uchar val;
    uchar page;
    uchar addr;

    PMPUTSTR("regs\n");
    for (page = 0; page < 4; ++page) {
        ISAout(0, (page << 6) | neCR_STP);
        val = ISAin(neCRrd);
        space();
        putint(page);
        putchar('/');
        putint(val >> 6);
        putchar(':');
        if (page != val >> 6) err = 1;
        for (addr = 0; addr < 0x10; ++addr) {
            if (page == 2 && addr > 2) break;
            if (page == 3 && addr > 0) break;
            val = ISAin(addr);
            space();
            puthexb(val);
        }
        crlf();
    }
    if (err) {
        PMPUTSTR("not ");
        NE2000();
        bang();
        crlf();
    }
}

void walk(
    char l1, char l2, char l3, char l4,
    uchar page1, uchar reg1,
    uchar page2, uchar reg2
){
    uchar i;
    uchar val;
    uchar expect;

    cr();
    putchar(l1);
    if (l2) putchar(l2);
    if (l3) putchar(l3);
    if (l4) putchar(l4);
    space();
    puthex(page1);
    putchar('/');
    puthex(reg1);
    crlf();

    for (i = 0; i < 9; ++i) {
        expect = (1 << i) & 0xff;
        progress(reg1, expect, 0);
        ISAout(0, (page1 << 6) | neCR_STP);
        ISAout(reg1, expect);
        ISAout(0, (page2 << 6) | neCR_STP);
        val = ISAin(reg2);
	if (val != expect) {
            oops(reg2, val, expect);
            //break;
        }
    }
    blank();
}

#define RAM_LOW   NE_RAM_OFFSET
#define RAM_HIGH  (NE_RAM_OFFSET + (NE_RAM_PAGES * 256))

void test(void)
{
    uchar val;
    uchar expect;
    ushort ram;

    led_on();

    crlf();
    NE2000();
    space();
    putchar('T');
    putchar('e');
    putchar('s');
    putchar('t');
    crlf();
    led_off();

    err = 0;

    ISAreset();
    ISAout(neRESETwr, 0xff);
    ISAout(neCRwr, neCR_STA | neCR_RD1 | neCR_STP); // page 0
    ISAout(neDCRwr, neDCR_LS | neDCR_FT1);

    show_regs();

    walk('B','N','R','Y', 0, neBNRYwr, 0, neBNRYrd);

    if (!err) {
        walk('P','A','R','0', 1, nePAR0wr, 1, nePAR0rd);
        walk('P','A','R','1', 1, nePAR1wr, 1, nePAR1rd);
        walk('P','A','R','2', 1, nePAR2wr, 1, nePAR2rd);
        walk('P','A','R','3', 1, nePAR3wr, 1, nePAR3rd);
        walk('P','A','R','4', 1, nePAR4wr, 1, nePAR4rd);
        walk('P','A','R','5', 1, nePAR5wr, 1, nePAR5rd);

        walk('C','U','R','R', 1, neCURRwr, 1, neCURRrd);

        walk('M','A','R','0', 1, neMAR0wr, 1, neMAR0rd);
        walk('M','A','R','1', 1, neMAR1wr, 1, neMAR1rd);
        walk('M','A','R','2', 1, neMAR2wr, 1, neMAR2rd);
        walk('M','A','R','3', 1, neMAR3wr, 1, neMAR3rd);
        walk('M','A','R','4', 1, neMAR4wr, 1, neMAR4rd);
        walk('M','A','R','5', 1, neMAR5wr, 1, neMAR5rd);
        walk('M','A','R','6', 1, neMAR6wr, 1, neMAR6rd);
        walk('M','A','R','7', 1, neMAR7wr, 1, neMAR7rd);

        walk('S','T','R','T', 0, nePSTARTwr, 2, nePSTARTrd);
        walk('S','T','O','P', 0, nePSTOPwr, 2, nePSTOPrd);
    }

    PMPUTSTR("DMA\n");
    expect = 0x00;
    for (ram = RAM_LOW; ram < RAM_HIGH; ++ram) {
        DMAout(ram, expect);
    }
    for (ram = RAM_LOW; ram < RAM_HIGH; ++ram) {
        val = DMAin(ram);
        progress(ram, expect, 0xff);
	if (val != expect) {
            oops(ram, val, expect);
            break;
        }
    }
    blank();

    PMPUTSTR("RAM\n");
    for (ram = RAM_LOW; ram < RAM_HIGH; ++ram) {
        expect = ram & 0xff;
        DMAout(ram, expect);
        progress(ram, val, 0xff);
        val = DMAin(ram);
	if (val != expect) {
            oops(ram, val, expect);
            break;
        }
        DMAout(ram, 0x00);
    }
    blank();

    show_regs();

    if (err) {
        Error();
        PMPUTSTR("s detected!!\n");
    } else {
        PMPUTSTR("No ");
        Error();
        PMPUTSTR("s\n");
    }

    wait();
    getchar();
    space();
    crlf();
}
